#include "stitcher.h"

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <limits>
#include <list>
#include <set>
#include <vector>

#ifdef WIN32
#define SIZE_T_FORMAT "%d"
#else
#define SIZE_T_FORMAT "%zd"
#endif

namespace simplestitcher {
using std::list;
using std::numeric_limits;
using std::set;
using std::swap;
using std::vector;

namespace {
const size_t kDescriptorLength = 128;
const float kDescriptorEpsilon = 1e-6;

size_t Match(const vector<float>& a_descriptors,
             const vector<float>& b_descriptors,
             float threshold,
             vector<int>* matches) {
  matches->clear();
  matches->resize(a_descriptors.size() / kDescriptorLength);

  assert(a_descriptors.size() % kDescriptorLength == 0);
  assert(b_descriptors.size() % kDescriptorLength == 0);

  size_t num_matches = 0;
  for (size_t i = 0; i < a_descriptors.size() / kDescriptorLength; i++) {
    float best = numeric_limits<float>::max();
    int best_match = -1;
    for (size_t j = 0; j < b_descriptors.size() / kDescriptorLength; j++) {
      float score = 0.0;
      // Split into chunks of four in the hopes that the compiler will
      // vectorize the operations.
      for (size_t k = 0; k < kDescriptorLength; k+= 4) {
        size_t a_idx = i * (kDescriptorLength) + k,
               b_idx = j * (kDescriptorLength) + k;
        double t0 = a_descriptors[a_idx] - b_descriptors[b_idx],
               t1 = a_descriptors[a_idx + 1] - b_descriptors[b_idx + 1],
               t2 = a_descriptors[a_idx + 2] - b_descriptors[b_idx + 2],
               t3 = a_descriptors[a_idx + 3] - b_descriptors[b_idx + 3];
        score += t0 * t0 + t1 * t1 + t2 * t2 + t3 * t3;
        if (score > best) break;
      }
      if (score < best) {
        best = score;
        best_match = j;
      }
    }
    if (best > threshold) {
      best_match = -1;
    }

    (*matches)[i] = static_cast<int>(best_match);
    if (best_match >= 0) {
      num_matches++;
    }
  }
  return num_matches;
}

template <typename T>
void Shuffle(vector<T>* v, size_t n) {
  size_t num = (n < v->size()) ? n : v->size();
  for (size_t i = 0; i < num; i++) {
    size_t j = rand() % (v->size() - i);
    swap((*v)[i], (*v)[j]);
  }
}

class Ransac {
 public:
  // a_points, b_points, and matches must outlive the Ransac object.
  Ransac(const vector<cv::KeyPoint>& a_points,
         const vector<cv::KeyPoint>& b_points,
         const vector<int>& matches,
         const RansacOptions& options)
      : options_(options),
        x_(0), y_(0), score_(numeric_limits<float>::max()),
        num_inliers_(0) {
    // Filter out bad matches.
    vector<int> valid_matches;
    for (size_t i = 0; i < matches.size(); i++) {
      if (matches[i] >= 0) {
        valid_matches.push_back(i);
      }
    }

    points_.resize(valid_matches.size());
    for (size_t i = 0; i < valid_matches.size(); i++) {
      points_[i].dx = a_points[valid_matches[i]].pt.x -
                      b_points[matches[valid_matches[i]]].pt.x;
      points_[i].dy = a_points[valid_matches[i]].pt.y -
                      b_points[matches[valid_matches[i]]].pt.y;
    }
  }

  float x() const {
    return x_;
  }

  float y() const {
    return y_;
  }

  float score() const {
    return score_;
  }

  size_t num_inliers() const {
    return num_inliers_;
  }

  void Run() {
    vector<PointMatch> inliers;
    for (size_t i = 0; i < options_.num_iterations; i++) {
      Shuffle(&points_, options_.sample_size);

      float px, py;
      AverageTranslation(points_, options_.sample_size, &px, &py);

      inliers.clear();
      FindInliers(px, py, &inliers);
      if (inliers.size() < options_.min_inliers) {
        continue;
      }

      float mx, my;
      AverageTranslation(inliers, inliers.size(), &mx, &my);
      float model_score = ScoreModel(mx, my, inliers);
      if (model_score < score_) {
        score_ = model_score;
        num_inliers_ = inliers.size();
        x_ = mx;
        y_ = my;
      }
    }
  }

 private:
  struct PointMatch {
    float dx, dy;
  };

  void GetTranslation(size_t idx, float *x, float *y) const {
    *x = points_[idx].dx;
    *y = points_[idx].dy;
  }

  void AverageTranslation(const vector<PointMatch>& points, size_t num_indices_to_use,
                          float *x, float *y) const {
    *x = 0.0f;
    *y = 0.0f;
    if (num_indices_to_use > points.size()) {
      num_indices_to_use = points.size();
    }
    for (size_t i = 0; i < num_indices_to_use; i++) {
      *x += points[i].dx;
      *y += points[i].dy;
    }
    *x /= static_cast<float>(num_indices_to_use);
    *y /= static_cast<float>(num_indices_to_use);
  }

  float ScoreModel(float x, float y, const vector<PointMatch>& points) const {
    float score = 0.0f;
    for (size_t i = 0; i < points.size(); i++) {
      const float dx = points[i].dx, dy = points[i].dy;
      score += (dx - x) * (dx - x) + (dy - y) * (dy - y);
    }
    return score / points.size();
  }

  void FindInliers(float x, float y, vector<PointMatch>* inliers) const {
    // Source is just indexes of shuffled good matches; the sample is taken
    // from the first sample_size_ indices of valid_matches_, so we want to
    // look over the remaining indices.
    for (size_t i = options_.sample_size; i < points_.size(); i++) {
      const float dx = points_[i].dx, dy = points_[i].dy;
      if ((dx - x) * (dx - x) + (dy - y) * (dy - y) <
          options_.inlier_tolerance) {
        inliers->push_back(points_[i]);
      }
    }
  }

  const RansacOptions options_;

  vector<PointMatch> points_;

  float x_;
  float y_;
  float score_;
  size_t num_inliers_;

  DISALLOW_COPY_AND_ASSIGN(Ransac);
};


class GlobalAlignment {
 public:
  GlobalAlignment() {}

  struct LocalAlignment {
    LocalAlignment(size_t reference_image, size_t aligned_image,
                   float x_offset, float y_offset, float score)
        : reference_image(reference_image),
          aligned_image(aligned_image),
          x_offset(x_offset),
          y_offset(y_offset),
          score(score) {}

    size_t reference_image;
    size_t aligned_image;
    float x_offset;
    float y_offset;
    float score;

    bool operator <(const LocalAlignment& other) const {
      return score < other.score;
    }
  };

  struct Alignment {
    Alignment(size_t image, float x_offset, float y_offset)
        : image(image), x_offset(x_offset), y_offset(y_offset) {}

    size_t image;
    float x_offset;
    float y_offset;
  };

  void AddLocalAlignment(const LocalAlignment& alignment) {
    alignments_.push_back(alignment);
  }

  const Alignment& alignment(size_t i) const {
    assert(i < global_alignments_.size() && global_alignments_[i] != NULL);
    return *global_alignments_[i];
  }

  bool Align() {
    set<size_t> unaligned_images;
    for (vector<LocalAlignment>::const_iterator i = alignments_.begin();
         i != alignments_.end(); ++i) {
      unaligned_images.insert(i->reference_image);
      unaligned_images.insert(i->aligned_image);
    }

    std::sort(alignments_.begin(), alignments_.end());

    list<LocalAlignment> local_alignments(alignments_.begin(), alignments_.end());

    // Effectively, we must compute a minimum spanning tree, where edges are
    // LocalAlignments and nodes are images.
    global_alignments_.clear();
    global_alignments_.resize(unaligned_images.size());

    // Start with the min-cost edge.
    list<LocalAlignment>::iterator first = local_alignments.begin();
    global_reference_image_ = first->reference_image;
    global_alignments_[first->reference_image] =
        new Alignment(first->reference_image, 0, 0);
    global_alignments_[first->aligned_image] = 
      new Alignment(first->aligned_image, first->x_offset,
                    first->y_offset);
    unaligned_images.erase(first->reference_image);
    unaligned_images.erase(first->aligned_image);
    fprintf(stderr, "  using reference alignment " SIZE_T_FORMAT
            " -> " SIZE_T_FORMAT "\n",
            first->reference_image, first->aligned_image);
    local_alignments.erase(first);

    // Then, until all nodes are in the tree (or all edges have been
    // exhausted), iteratively add the cheapest edge connecting the tree to
    // new node.
    while (!unaligned_images.empty() && !local_alignments.empty()) {
      bool added_any = false;
      for (list<LocalAlignment>::iterator i = local_alignments.begin();
           i != local_alignments.end(); ++i) {
        const LocalAlignment& alignment = *i;
        if (ContainsKey(unaligned_images, alignment.aligned_image) &&
            !ContainsKey(unaligned_images, alignment.reference_image)) {
          assert(global_alignments_[alignment.reference_image] != NULL);
          const Alignment& reference_alignment =
              *global_alignments_[alignment.reference_image];
          fprintf(stderr, "  adding local alignment "
                  SIZE_T_FORMAT " -> " SIZE_T_FORMAT "\n",
                  alignment.reference_image,
                  alignment.aligned_image);
          global_alignments_[alignment.aligned_image] = new Alignment(
              alignment.aligned_image,
              alignment.x_offset + reference_alignment.x_offset,
              alignment.y_offset + reference_alignment.y_offset);
          unaligned_images.erase(alignment.aligned_image);
          local_alignments.erase(i);
          added_any = true;
          break;
        } else if (ContainsKey(unaligned_images, alignment.reference_image) &&
                   !ContainsKey(unaligned_images, alignment.aligned_image)) {
          assert(global_alignments_[alignment.aligned_image] != NULL);
          const Alignment& reference_alignment =
              *global_alignments_[alignment.aligned_image];
          fprintf(stderr, "  adding local alignment "
                  SIZE_T_FORMAT " -> " SIZE_T_FORMAT "\n",
                  alignment.aligned_image,
                  alignment.reference_image);
          global_alignments_[alignment.reference_image] = new Alignment(
              alignment.reference_image,
              -alignment.x_offset + reference_alignment.x_offset,
              -alignment.y_offset + reference_alignment.y_offset);
          unaligned_images.erase(alignment.reference_image);
          local_alignments.erase(i);
          added_any = true;
          break;
        }
      }

      if (!added_any) {
        break;
      }
    }

    if (!unaligned_images.empty()) {
      fprintf(stderr, "Unable to compute a global alignment; the following "
              "images are not connected: ");
      for (set<size_t>::const_iterator i = unaligned_images.begin();
           i != unaligned_images.end(); ++i) {
        fprintf(stderr, " " SIZE_T_FORMAT, *i);
      }
      fprintf(stderr, "\n");
      return false;
    }

    float min_x = numeric_limits<float>::max(),
          min_y = numeric_limits<float>::max();
    for (size_t i = 0; i < global_alignments_.size(); i++) {
      assert(global_alignments_[i] != NULL);
      const Alignment& a = *global_alignments_[i];
      if (a.x_offset < min_x) min_x = a.x_offset;
      if (a.y_offset < min_y) min_y = a.y_offset;
    }

    fprintf(stderr, "Computed global alignment (using "
            SIZE_T_FORMAT " as reference image):\n",
            global_reference_image_);
    for (size_t i = 0; i < global_alignments_.size(); i++) {
      global_alignments_[i]->x_offset -= min_x;
      global_alignments_[i]->y_offset -= min_y;
      fprintf(stderr, "  " SIZE_T_FORMAT ": (%f, %f)\n",
              global_alignments_[i]->image,
              global_alignments_[i]->x_offset,
              global_alignments_[i]->y_offset);
    }

    return true;
  }

 private:
  vector<LocalAlignment> alignments_;
  vector<Alignment*> global_alignments_;
  size_t global_reference_image_;

  DISALLOW_COPY_AND_ASSIGN(GlobalAlignment);
};

unsigned char Sample(const cv::Mat& image, float x, float y, int channel) {
  float tx = x - floor(x);
  float ty = y - floor(y);

  float v[4][4];

  int x0 = floor(x);
  int y0 = floor(y);
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      int cx = x0 + i - 1;
      int cy = y0 + j - 1;

      if (cx > 0 && cx < static_cast<int>(image.cols) &&
          cy > 0 && cy < static_cast<int>(image.rows)) {
        v[i][j] = static_cast<float>(
            image.ptr<unsigned char>(cy)[3 * cx + channel]);
      } else {
        v[i][j] = 0;
      }
    }
  }

  float xi[4], r;

  xi[0] = 0.5 * ((    v[3][0] - 3 * v[2][0] + 3 * v[1][0] -     v[0][0]) * tx * tx * tx +
                 (   -v[3][0] + 4 * v[2][0] - 5 * v[1][0] + 2 * v[0][0]) * tx * tx +
                 (                  v[2][0] -                   v[0][0]) * tx +
                 (                            2 * v[1][0]              )                 );
  xi[1] = 0.5 * ((    v[3][1] - 3 * v[2][1] + 3 * v[1][1] -     v[0][1]) * tx * tx * tx +
                 (   -v[3][1] + 4 * v[2][1] - 5 * v[1][1] + 2 * v[0][1]) * tx * tx +
                 (                  v[2][1] -                   v[0][1]) * tx +
                 (                            2 * v[1][1]              )                 );
  xi[2] = 0.5 * ((    v[3][2] - 3 * v[2][2] + 3 * v[1][2] -     v[0][2]) * tx * tx * tx +
                 (   -v[3][2] + 4 * v[2][2] - 5 * v[1][2] + 2 * v[0][2]) * tx * tx +
                 (                  v[2][2] -                   v[0][2]) * tx +
                 (                            2 * v[1][2]              )                 );
  xi[3] = 0.5 * ((    v[3][3] - 3 * v[2][3] + 3 * v[1][3] -     v[0][3]) * tx * tx * tx +
                 (   -v[3][3] + 4 * v[2][3] - 5 * v[1][3] + 2 * v[0][3]) * tx * tx +
                 (                  v[2][3] -                   v[0][3]) * tx +
                 (                            2 * v[1][3]              )                 );
  r     = 0.5 * ((    xi[3] - 3 * xi[2] + 3 * xi[1] -     xi[0]) * ty * ty * ty +
                 (   -xi[3] + 4 * xi[2] - 5 * xi[1] + 2 * xi[0]) * ty * ty +
                 (                xi[2] -                 xi[0]) * ty +
                 (                        2 * xi[1]            )                 );

  return (r < 0) ? 0 : ((r > 255) ? 255 : static_cast<unsigned char>(r));
}
}  // namespace


void Stitcher::AddImage(const cv::Mat& image) {
  assert(image.type() == CV_8UC3);
  images_.push_back(image);
}

bool Stitcher::AddImageFromFile(const std::string& filename) {
  cv::Mat image = cv::imread(filename, 1 /* force RGB */);
  if (image.data == NULL) {
    fprintf(stderr, "Error reading image \"%s\".\n", filename.c_str());
    return false;
  }
  fprintf(stderr, "Read image \"%s\".\n", filename.c_str());
  AddImage(image);
  return true;
}

bool Stitcher::Stitch() {
  cv::SURF detector;
  vector<vector<cv::KeyPoint> > keypoints(images_.size());
  vector<vector<float> > descriptors(images_.size());
  for (size_t i = 0; i < images_.size(); i++) {
    cv::Mat image;
    cv::cvtColor(images_[i], image, CV_RGB2GRAY);

    cv::Mat empty_mask(image.rows, image.cols, CV_8UC1, 0xFF);
    detector(image, empty_mask, keypoints[i], descriptors[i]);
    fprintf(stderr, "(" SIZE_T_FORMAT " of " SIZE_T_FORMAT ") extracted "
            SIZE_T_FORMAT " keypoints, with " SIZE_T_FORMAT " descriptor "
            "values.\n",
            i + 1, images_.size(),
            keypoints[i].size(),
            descriptors[i].size());
  }

  fprintf(stderr, "Computing pair alignments\n");
  GlobalAlignment g;
  for (size_t i = 0; i < images_.size(); i++) {
    for (size_t j = i + 1; j < images_.size(); j++) {
      vector<int> matches;
      size_t num_matches = Match(
          descriptors[i],
          descriptors[j],
          options_.descriptor_match_threshold,
          &matches);
      fprintf(stderr, "  " SIZE_T_FORMAT " x " SIZE_T_FORMAT ": "
              SIZE_T_FORMAT " matches\n", i, j, num_matches);


      Ransac r(keypoints[i], keypoints[j], matches, options_.ransac_options);
      r.Run();
      if (r.num_inliers() > 0) {
        fprintf(stderr, "    -> RANSAC: (%f, %f) %f (" SIZE_T_FORMAT " inliers)\n",
                r.x(), r.y(), r.score(), r.num_inliers());
        g.AddLocalAlignment(GlobalAlignment::LocalAlignment(
            i, j, r.x(), r.y(), r.score()));
      } else {
        fprintf(stderr, "    -> RANSAC failed.\n");
      }
    }
  }

  fprintf(stderr, "Computing global alignment\n");
  if (!g.Align()) {
    return false;
  }

  float max_x = 0.0f, max_y = 0.0f;
  for (size_t i = 0; i < images_.size(); i++) {
    float x_right = g.alignment(i).x_offset + images_[i].cols;
    float y_bottom = g.alignment(i).y_offset + images_[i].rows;

    if (x_right > max_x) max_x = x_right;
    if (y_bottom > max_y) max_y = y_bottom;
  }

  size_t cols = ceil(max_x), rows = ceil(max_y);
  output_ = cv::Mat(rows, cols, CV_8UC3);
  output_.setTo(0);
  fprintf(stderr, "Aligned image has dimensions "
          SIZE_T_FORMAT " x " SIZE_T_FORMAT ".\n", cols, rows);
  for (size_t y = 0; y < rows; y++) {
    for (size_t x = 0; x < cols; x++) {
      for (size_t i = 0; i < images_.size(); i++) {
        float ix = x - g.alignment(i).x_offset,
              iy = y - g.alignment(i).y_offset;
        if (ix > 1 && iy > 1 && ix < images_[i].cols - 1 && iy < images_[i].rows - 1) {
          unsigned char *pr = &(output_.ptr<unsigned char>(y)[x * 3 + 0]),
                        *pg = &(output_.ptr<unsigned char>(y)[x * 3 + 1]),
                        *pb = &(output_.ptr<unsigned char>(y)[x * 3 + 2]);
          unsigned char r = Sample(images_[i], ix, iy, 0),
                        g = Sample(images_[i], ix, iy, 1),
                        b = Sample(images_[i], ix, iy, 2);
          if (r > *pr || g > *pg || b > *pb) {
            *pr = r;
            *pg = g;
            *pb = b;
            break;
          }
        }
      }
    }
  }

  return true;
}
}  // namespace simplestitcher
